# -*- coding: utf-8 -*-
import bisect
import logging
import re
import sys
import time
import yaml
from functools import lru_cache
from os.path import isfile

logger = logging.getLogger(__name__)


def ip2long(ip):
    "convert decimal dotted quad string to long integer"
    hexn = ''.join(["%02X" % int(i) for i in ip.split('.')])
    return int(hexn, 16)


def long2ip(num):
    "convert long int to dotted quad string"
    iplist = []
    for _ in range(4):
        num, mod = divmod(num, 256)
        iplist.insert(0, str(mod))
    return '.'.join(iplist)


class BaseFinder(object):

    def __init__(self, ipfile, recordfile):
        if not isfile(ipfile) or not isfile(recordfile):
            logger.warning("can't find ip data file: %s" % ipfile)
        else:
            self.ipfile = ipfile
            self.recordfile = recordfile

            # 初始化iplist，用来进行2分查找
            self.iplist = []
            # 初始化iphash，用来检索详细信息
            self.iphash = {}

            # 初始化存储a.yaml配置
            self.record = {}
            # 存储各个域名的地域对于ip信息
            self.locmapip = {}

            # load record data
            self.loadRecord()

            # load ip data
            self.loadIP()

            print('Init IP pool finished !')

    def loadIP(self):
        f = open(self.ipfile, 'r')
        logger.warning("before load: %s" % (time.time()))
        for eachline in f:
            ipstart, ipend, country, province, city, sp = eachline.strip().split(',')
            ipstart = int(ipstart)
            ipend = int(ipend)

            # 如果ip地址为0,忽略
            if 0 == ipstart:
                continue
            self.iplist.append(ipstart)
            if ipstart in self.iphash:
                # print "maybe has same ipstart"
                pass
            else:
                # ipstart, ipend, country, province, city, sp, domain ip hash
                self.iphash[ipstart] = [ipstart, ipend,
                                        country, province, city, sp, {}]
                # 最好合并后再计算
                self.joinIP(ipstart)
        f.close()
        logger.warning("after load: %s" % (time.time()))
        self.iplist.sort()
        logger.warning("after sort: %s" % (time.time()))

    # 重写LoadRecord和JoinIP，提升启动效率
    def loadRecord(self):
        Add = [8, 4, 2, 1]
        f = open(self.recordfile, 'r')
        self.record = yaml.load(f, Loader=yaml.FullLoader)
        for fqdn in self.record:
            self.locmapip[fqdn] = {}
            if fqdn.endswith("_template"):
                continue

            for router in self.record[fqdn]:
                if router == 'default' or router == 'ttl':
                    continue
                p = None
                # p = re.match(ur"(.*),(.*),(.*),(.*)", router)
                p = router.strip().split(',')
                if p is None:
                    logger.warning(
                        "maybe record file format error: %s" % self.recordfile)
                    sys.exit(1)
                match = [None, None, None, None]
                weight = 0
                for num in range(0, 4):
                    match[num] = p[num]
                    if p[num] != "":
                        weight += Add[num]

                if match[0] not in self.locmapip[fqdn]:
                    self.locmapip[fqdn][match[0]] = {}
                    self.locmapip[fqdn][match[0]][match[1]] = {}
                    self.locmapip[fqdn][match[0]][match[1]][match[2]] = {}
                    self.locmapip[fqdn][match[0]][match[1]][match[2]][match[3]] = \
                        [self.record[fqdn][router], weight]
                elif match[1] not in self.locmapip[fqdn][match[0]]:
                    self.locmapip[fqdn][match[0]][match[1]] = {}
                    self.locmapip[fqdn][match[0]][match[1]][match[2]] = {}
                    self.locmapip[fqdn][match[0]][match[1]][match[2]][match[3]] = \
                        [self.record[fqdn][router], weight]
                elif match[2] not in self.locmapip[fqdn][match[0]][match[1]]:
                    self.locmapip[fqdn][match[0]][match[1]][match[2]] = {}
                    self.locmapip[fqdn][match[0]][match[1]][match[2]][match[3]] = \
                        [self.record[fqdn][router], weight]
                elif match[3] not in self.locmapip[fqdn][match[0]][match[1]][match[2]]:
                    self.locmapip[fqdn][match[0]][match[1]][match[2]][match[3]] = \
                        [self.record[fqdn][router], weight]
        f.close()
        # logger.warning(self.locmapip)

    def joinIP(self, ip):
        for fqdnk, fqdnv in self.locmapip.items():
            l1 = []
            l2 = []
            l3 = []
            weight = 0
            # logger.warning("l1 : %s, %s" %(self.iphash[ip][2], fqdnv.keys()))
            if "" in fqdnv and "" != self.iphash[ip][2]:
                l1.append(fqdnv[""])
            if self.iphash[ip][2] in fqdnv:
                l1.append(fqdnv[self.iphash[ip][2]])
            for k in l1:
                # logger.warning("l2 : %s, %s" %(self.iphash[ip][3], k.keys()))
                if "" in k and "" != self.iphash[ip][3]:
                    l2.append(k[""])
                if self.iphash[ip][3] in k:
                    l2.append(k[self.iphash[ip][3]])
            for k in l2:
                # logger.warning("l3 : %s, %s" %(self.iphash[ip][4], k.keys()))
                if "" in k and "" != self.iphash[ip][4]:
                    l3.append(k[""])
                if self.iphash[ip][4] in k:
                    l3.append(k[self.iphash[ip][4]])
            for k in l3:
                # logger.warning("l4 : %s, %s" %(self.iphash[ip][5], k.keys()))
                if "" in k and k[""][1] > weight:
                    self.iphash[ip][6][fqdnk] = k[""]
                    weight = k[""][1]
                if self.iphash[ip][5] in k and k[self.iphash[ip][5]][1] > weight:
                    self.iphash[ip][6][fqdnk] = k[self.iphash[ip][5]]
                    weight = k[self.iphash[ip][5]][1]
            if fqdnk not in self.iphash[ip][6]:
                self.iphash[ip][6][fqdnk] = [self.record[fqdnk]['default'], 0]

    def listIP(self):
        for key in self.iphash:
            print("ipstart: %s  ipend: %s  country: %s  province: %s  city: %s  sp: %s" % (
                key, self.iphash[key][1], self.iphash[key][2], self.iphash[key][3], self.iphash[key][4],
                self.iphash[key][5]))
            for i in self.iphash[key][6]:
                print("[domain:%s   ip: %s]" % (i, self.iphash[key][6][i][0]))

    def searchLocation(self, ip):
        ipnum = ip2long(ip)
        ip_point = bisect.bisect_right(self.iplist, ipnum)
        i = self.iplist[ip_point - 1]
        if ip_point == self.iplist.__len__():
            j = self.iplist[ip_point - 1]
        else:
            j = self.iplist[ip_point]

        return i, j, ipnum

    @lru_cache(maxsize=2048 * 2048, typed=True)
    def findIP(self, ip, name):
        i, _, ipnum = self.searchLocation(ip)
        ip_list = None
        if i in self.iphash:
            ipstart = self.iphash[i][0]
            ipend = self.iphash[i][1]
            country = self.iphash[i][2]
            province = self.iphash[i][3]
            city = self.iphash[i][4]
            sp = self.iphash[i][5]
            if ipstart <= ipnum <= ipend:
                ip_list = [
                    tmp_ip for tmp_ip in re.split(r',|\s+', self.iphash[i][6][name][0])
                    if not re.search(r'[^0-9.]', tmp_ip)]
                logger.info("userip:[%s] domain:[%s] section:[%s-%s] location:[%s,%s,%s,%s] ip_list:%s" % (
                    ip, name, long2ip(ipstart), long2ip(ipend), country, province, city, sp, ip_list))
        if not ip_list or len(ip_list) == 0:
            # maybe something wrong
            ip_list = [
                tmp_ip for tmp_ip in re.split(r',|\s+', self.record[name]['default'])
                if not re.search(r'[^0-9.]', tmp_ip)]
        return ip_list


class Finder(object):

    def __init__(self, ipfile, recordfile, monitor_mapping):
        self.monitor_mapping = monitor_mapping
        self.finder = BaseFinder(ipfile, recordfile)

    def findIP(self, ip, name):
        tmp_ip_list = self.finder.findIP(ip, name)
        ip_list = [
            tmp_ip for tmp_ip in tmp_ip_list if self.monitor_mapping.check(name, tmp_ip)]
        if len(ip_list) == 0:
            logger.warning("no available ip for %s, use default ip" % name)
            return tmp_ip_list
        return ip_list



# -*- coding: utf-8 -*-

# version
__version__ = '1.0.1.1'

import logging
import os
import sys
import yaml
from multiprocessing import cpu_count, Process
from os.path import isfile
from twisted.internet import reactor
from twisted.names import dns

from . import server, monitor
from .finder import Finder

logger = logging.getLogger(__name__)


def loadconfig(path):
    if not isfile(path):
        print("[FATAL] can't find config file %s !" % path)
        exit(1)
    with open(path, 'r') as f:
        return yaml.load(f, Loader=yaml.FullLoader)


def prepare_run(run_env):
    # load main config
    logger.info('start to load %s ......' % run_env['conf'])
    conf = loadconfig(os.path.join(run_env['conf'], 'sdns.yaml'))
    logger.info('start to load A,SOA,NS record ......')
    a_mapping = loadconfig(os.path.join(run_env['conf'], 'a.yaml'))
    ns_mapping = loadconfig(os.path.join(run_env['conf'], 'ns.yaml'))
    soa_mapping = loadconfig(os.path.join(run_env['conf'], 'soa.yaml'))

    # start monitor
    monitor_config = loadconfig(os.path.join(run_env['conf'], 'monitor.yaml'))
    monitor_mapping = monitor.MonitorMapping(monitor_config, a_mapping)
    # load dns record config file
    logger.info('start to init IP pool ......')
    finder = Finder(
        os.path.join(run_env['conf'], 'ip.csv'),
        os.path.join(run_env['conf'], 'a.yaml'),
        monitor_mapping)

    run_env['finder'] = finder

    # set up a resolver that uses the mapping or a secondary nameserver
    dns_forwards = []
    for i in conf['dns_forwards']:
        dns_forward_ip, dns_forward_port = i.split(":")
        dns_forwards.append((dns_forward_ip, int(dns_forward_port)))

    # create the protocols
    for listen_tcp in conf['listen']['tcp']:
        listen_tcp_ip, listen_tcp_port = listen_tcp.split(":")
        f = server.SmartDNSFactory(clients=[
            server.MapResolver(
                finder, a_mapping, ns_mapping, soa_mapping, servers=dns_forwards)])
        f.noisy = False
        run_env['tcp'].append([int(listen_tcp_port), f, listen_tcp_ip])
    for listen_udp in conf['listen']['tcp']:
        listen_udp_ip, listen_udp_port = listen_udp.split(":")
        p = dns.DNSDatagramProtocol(server.SmartDNSFactory(clients=[
            server.MapResolver(
                finder, a_mapping, ns_mapping, soa_mapping, servers=dns_forwards)]))
        p.noisy = False
        run_env['udp'].append([int(listen_udp_port), p, listen_udp_ip])
    return conf


def main():
    run_env = {'udp': [], 'tcp': [], 'closed': 0,
               'updated': False, 'finder': None}
    if len(sys.argv) > 1:
        run_env['conf'] = sys.argv[1]
    else:
        run_env['conf'] = '/etc/smartdns'

    conf = prepare_run(run_env)
    for e in run_env['tcp']:
        reactor.listenTCP(e[0], e[1], interface=e[2])
    for e in run_env['udp']:
        reactor.listenUDP(e[0], e[1], interface=e[2])
    workers = []
    for _ in range(conf.get("workers", cpu_count() * 2)):
        process = Process(target=reactor.run)
        process.daemon = True
        process.start()
        workers.append(process)
    [worker.join() for worker in workers]



from twisted.internet import task, ssl, reactor
from twisted.web.client import Agent, BrowserLikePolicyForHTTPS
from twisted.web.http_headers import Headers
from twisted.web.iweb import IPolicyForHTTPS
from urllib.parse import urlparse
from zope.interface import implementer


@implementer(IPolicyForHTTPS)
class SmartClientContextFactory(object):

    def __init__(self):
        self.default_policy = BrowserLikePolicyForHTTPS

    def creatorForNetloc(self, hostname, port):
        return ssl.CertificateOptions(verify=False)


class Monitor(object):

    def __init__(self, ip_set, monitor):
        self.ip_set = ip_set
        self.monitor = monitor
        self.black_mapping = {}

    def _check(self):
        host = urlparse(self.monitor["url"]).netloc
        for ip in self.ip_set:
            if ip not in self.black_mapping:
                self.black_mapping[ip] = 0
            url = self.monitor['url'].replace(host, ip, 1).encode("utf8")
            agent = Agent(reactor, contextFactory=SmartClientContextFactory(), connectTimeout=30)
            agent.request(b'GET', url, headers=Headers({"host": [host, ]})).addCallbacks(
                BlackMappingRemover(ip, self.black_mapping), BlackMappingAdder(ip, self.black_mapping))

    def check(self, ip):
        return self.black_mapping[ip] < self.monitor["frequency"]

    def start(self):
        task.LoopingCall(self._check).start(self.monitor["interval"])


class BlackMappingRemover(object):

    def __init__(self, ip, black_mapping):
        self.ip = ip
        self.black_mapping = black_mapping

    def __call__(self, *args, **kwargs):
        self.black_mapping[self.ip] = 0


class BlackMappingAdder(object):

    def __init__(self, ip, black_mapping):
        self.ip = ip
        self.black_mapping = black_mapping

    def __call__(self, *args, **kwargs):
        self.black_mapping[self.ip] += 1


class MonitorMapping(object):

    def __init__(self, config, amapping):
        self.config = config
        self.monitor_mapping = self._make_monitor_mapping(amapping)

    def _make_monitor_mapping(self, amapping):
        monitor_mapping = {}
        for name, item in amapping.items():
            if name in self.config:
                ip_set = set()
                monitor = Monitor(ip_set, self.config[name])
                for key, value in item.items():
                    if key == 'ttl':
                        continue
                    else:
                        ip_set.update(value.split(' '))
                monitor.start()
                monitor_mapping[name] = monitor
        return monitor_mapping

    def check(self, name, ip):
        if name in self.monitor_mapping:
            return self.monitor_mapping[name].check(ip)
        return True



#coding:utf-8

import os
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(here, 'README.md'), encoding="utf8") as f:
    README = f.read()

requires = [
    'twisted',
    'pyyaml',
    'pyopenssl',
    'setuptools',
    'service_identity',
]

test_requirements = [
    'nose',
]


setup(

    name='smartdns',
    description='An Intelligent DNS Similar to DNSPod',
    version='3.3.3',
    author='duanhongyi',
    author_email='duanhyi@gmail.com',
    packages=find_packages(),
    include_package_data=True,
    long_description=README,
    url='https://github.com/duanhongyi/smartdns',
    install_requires=requires,
    tests_require=test_requirements,
    platforms='all platform',
    license='BSD',
    long_description_content_type='text/markdown',
    # install
    entry_points={'console_scripts': ['sdns=smartdns.sdns:main']}
)



# -*- coding: utf-8 -*-
import logging
import random
import re
from twisted.internet import defer
from twisted.names import dns, server, client, common, resolve
from twisted.python import failure

logger = logging.getLogger(__name__)

typeToMethod = {
    dns.A: 'lookupAddress',
    dns.AAAA: 'lookupIPV6Address',
    dns.A6: 'lookupAddress6',
    dns.NS: 'lookupNameservers',
    dns.CNAME: 'lookupCanonicalName',
    dns.SOA: 'lookupAuthority',
    dns.MB: 'lookupMailBox',
    dns.MG: 'lookupMailGroup',
    dns.MR: 'lookupMailRename',
    dns.NULL: 'lookupNull',
    dns.WKS: 'lookupWellKnownServices',
    dns.PTR: 'lookupPointer',
    dns.HINFO: 'lookupHostInfo',
    dns.MINFO: 'lookupMailboxInfo',
    dns.MX: 'lookupMailExchange',
    dns.TXT: 'lookupText',
    dns.SPF: 'lookupSenderPolicy',

    dns.RP: 'lookupResponsibility',
    dns.AFSDB: 'lookupAFSDatabase',
    dns.SRV: 'lookupService',
    dns.NAPTR: 'lookupNamingAuthorityPointer',
    dns.AXFR: 'lookupZone',
    dns.ALL_RECORDS: 'lookupAllRecords',
}

smartType = ('lookupAddress', 'lookupAuthority')


class FailureHandler:
    def __init__(self, resolver, query, timeout, addr=None, edns=None):
        self.resolver = resolver
        self.query = query
        self.timeout = timeout
        self.addr = addr
        self.edns = edns

    def __call__(self, failure):
        # AuthoritativeDomainErrors should halt resolution attempts
        failure.trap(dns.DomainError, defer.TimeoutError, NotImplementedError)
        return self.resolver(self.query, self.timeout, self.addr, self.edns)


class MapResolver(client.Resolver):
    def __init__(self, finder, a_mapping, ns_mapping, soa_mapping, servers):
        self.finder = finder
        self.a_mapping = a_mapping
        self.ns_mapping = ns_mapping
        self.soa_mapping = soa_mapping
        client.Resolver.__init__(self, servers=servers)

    def query(self, query, timeout=None, addr=None, edns=None):
        try:
            if typeToMethod[query.type] in smartType:
                return self.typeToMethod[query.type](str(query.name), timeout, addr, edns)
            else:
                return self.typeToMethod[query.type](str(query.name), timeout)
        except KeyError as e:
            return defer.fail(failure.Failure(NotImplementedError(str(self.__class__) + " " + str(query.type))))

    def lookupAddress(self, name, timeout=None, addr=None, edns=None):
        if name in self.a_mapping:
            ttl = self.a_mapping[name]['ttl']

            def packResult(value):
                ret = []
                add = []
                for x in value:
                    ret.append(dns.RRHeader(name, dns.A, dns.IN, ttl, dns.Record_A(x, ttl), True))

                if edns is not None:
                    if edns.rdlength > 8:
                        add.append(dns.RRHeader('', dns.EDNS, 4096, edns.ttl, edns.payload, True))
                    else:
                        add.append(dns.RRHeader('', dns.EDNS, 4096, 0, dns.Record_EDNS(None, 0), True))
                return [ret, (), add]

            result = self.finder.findIP(str(addr[0]), name)
            # 返回的IP数组乱序
            random.shuffle(result)
            return packResult(result)
        else:
            return self._lookup(name, dns.IN, dns.A, timeout)

    def lookupNameservers(self, name, timeout=None):
        if name in self.ns_mapping:
            result = self.ns_mapping[name]
            ttl = result['ttl']
            record = re.split(r',|\s+', result['record'])

            def packResultNS(value):
                ret = []
                for x in value:
                    ret.append(dns.RRHeader(name, dns.NS, dns.IN, ttl, dns.Record_NS(x, ttl), True))
                return [ret, (), ()]

            return packResultNS(record)
        else:
            return self._lookup(name, dns.IN, dns.NS, timeout)

    def lookupAuthority(self, name, timeout=None, addr=None, edns=None):
        if name in self.soa_mapping:
            result = self.soa_mapping[name]
            add = []

            def packResultSOA(value):
                if edns is not None:
                    if edns.rdlength > 8:
                        add.append(dns.RRHeader('', dns.EDNS, 4096, edns.ttl, edns.payload, True))
                    else:
                        add.append(dns.RRHeader('', dns.EDNS, 4096, 0, dns.Record_EDNS(None, 0), True))

                return [(dns.RRHeader(name, dns.SOA, dns.IN, value['ttl'],
                                      dns.Record_SOA(value['record'], value['email'], value['serial'], value['refresh'],
                                                     value['retry'], value['expire'], value['ttl']), True),),
                        (),
                        add
                        ]

            ret = packResultSOA(result)
            logger.info("SOA\t[domain: %s]\t[return: %s]\t[additional: %s]" % \
                        (name, result, add))
            return ret
        else:
            return self._lookup(name, dns.IN, dns.SOA, timeout)

    def lookupIPV6Address(self, name, timeout=None, addr=None):
        return [(), (), ()]


class SmartResolverChain(resolve.ResolverChain):

    def __init__(self, resolvers):
        # resolve.ResolverChain.__init__(self, resolvers)
        common.ResolverBase.__init__(self)
        self.resolvers = resolvers

    def _lookup(self, name, cls, type, timeout, addr=None, edns=None):
        q = dns.Query(name, type, cls)
        d = defer.fail(failure.Failure(dns.DomainError(name)))
        for r in self.resolvers:
            d = d.addErrback(
                FailureHandler(r.query, q, timeout, addr, edns)
            )
        return d

    def _query(self, query, timeout=None, addr=None, edns=None):
        if typeToMethod[query.type] in smartType:
            return self.typeToMethod[query.type](str(query.name), timeout, addr, edns)
        else:
            return self.typeToMethod[query.type](str(query.name), timeout)

    def query(self, query, timeout=None, addr=None, edns=None):
        try:
            return self._query(query, timeout, addr, edns)
        except KeyError as e:
            return defer.fail(failure.Failure(NotImplementedError(str(self.__class__) + " " + str(query.type))))

    def lookupAddress(self, name, timeout=None, addr=None, edns=None):
        return self._lookup(name, dns.IN, dns.A, timeout, addr, edns)

    def lookupAuthority(self, name, timeout=None, addr=None, edns=None):
        return self._lookup(name, dns.IN, dns.SOA, timeout, addr, edns)

    def lookupIPV6Address(self, name, timeout=None, addr=None, edns=None):
        return self._lookup(name, dns.IN, dns.AAAA, timeout, addr, edns)

    def lookupNameservers(self, name, timeout=None, addr=None, edns=None):
        return self._lookup(name, dns.IN, dns.NS, timeout, addr, edns)


class SmartDNSFactory(server.DNSServerFactory):

    def handleQuery(self, message, protocol, address):

        # if len(message.additional) > 0:
        #    print inspect.getmembers(message.additional[0]
        # 可以支持多个query
        query = message.queries[0]
        edns = None
        cliAddr = address
        if query.type == 43 or typeToMethod[query.type] == 'lookupAllRecords':
            return [(), (), ()]
        if isinstance(protocol, dns.DNSProtocol):
            cliAddr = protocol.transport.client
        elif typeToMethod[query.type] in smartType and \
                len(message.additional) != 0 and \
                message.additional[0].type == 41 and \
                message.additional[0].rdlength > 8:
            if isinstacne(message.additional[0].payload, dns.Record_A):
                cliAddr = (message.additional[0].payload.dottedQuad(), 0)
            edns = message.additional[0]
        return self.resolver.query(query, addr=cliAddr, edns=edns).addCallback(
            self.gotResolverResponse, protocol, message, address
        ).addErrback(
            self.gotResolverError, protocol, message, address
        )

    def __init__(self, authorities=None, clients=None, verbose=0):
        resolvers = []
        if authorities is not None:
            resolvers.extend(authorities)
        if clients is not None:
            resolvers.extend(clients)

        self.canRecurse = not not clients
        self.resolver = SmartResolverChain(resolvers)
        self.verbose = verbose
        self.connections = []
